from argparse import SUPPRESS, HelpFormatter, _SubParsersAction, _HelpAction
from collections import OrderedDict
import datetime
import os
import time
import re


DEFAULT_GROUP_NAMES = {
    # We replace ArgumentGroup title (value) with alias (key).
    None: [
        'positional arguments',
    ],
    'OPTIONS': [
        'optional arguments',
        'options',
    ]
}


DEFAULT_GROUP_NAMES_SUBCOMMANDS = {
    # We replace ArgumentGroup title (value) with alias (key).
    "arguments:": [
        'positional arguments',
    ],
    'options:': [
        'optional arguments',
        'options',
    ]
}


# all manpage attributes that can be set via CLI or setup.cfg
MANPAGE_DATA_ATTRS = (
    "authors",  # just "author" in setup.cfg, can be specified multiple times
    "description",
    "long_description",  # not in the manpage.py, we use parser.description
    "project_name",  # maps to distribution.get_name()
    "prog",
    "url",
    "version",
    "format",
    "manual_section",
    "manual_title",
    "include",
    "manfile",
)

# manpage sections that are handled specially, so need special treatment
# when --include'ing extra material; see Manpage.add_section.
SPECIAL_MANPAGE_SECTIONS = (
    "author",
    "comments",
    "description",
    "distribution",
    "synopsis",
)


def _markup(text):
    """
    Escape the text for the markdown format.
    """
    if isinstance(text, str):
        return text.replace('\\', '\\\\').replace('-', r'\-')
    return text


def get_manpage_data_from_distribution(distribution, data):
    """
    Update `data` with values from `distribution`.
    """
    # authors
    if not "authors" in data:
        author = None
        if distribution.get_author():
            author = distribution.get_author()
            if distribution.get_author_email():
                author += " <{}>".format(distribution.get_author_email())
        elif distribution.get_author_email():
            author = distribution.get_author_email()
        if author:
            data["authors"] = [author]

    attrs = list(MANPAGE_DATA_ATTRS)
    attrs.remove("authors")
    attrs.remove("prog")  # not available, copied from 'project_name' later
    attrs.remove("format")  # not available, must be set in setup.cfg
    # we want the utility description, not the project description
    attrs.remove("description")
    for attr in attrs:
        if data.get(attr, None):
            continue

        # map data["project_name"] to distribution.get_name()
        get_attr = "name" if attr == "project_name" else attr

        getter = getattr(distribution, "get_" + get_attr, None)
        if not getter:
            continue
        value = getter()

        data[attr] = value

    if "prog" not in data:
        data["prog"] = data["project_name"]


def _get_footer_lines(data):
    ret = []
    project_name = data.get("project_name", "")
    authors = data.get("authors")
    url = data.get("url")

    needs_separator = False
    if authors:
        ret.append('.SH AUTHOR')
        for author in authors:
            ret.append(".nf")
            ret.append(author)
            ret.append(".fi")
        needs_separator = True

    if url:
        if needs_separator:
            ret.append("")
        ret.append(".SH DISTRIBUTION")
        ret.append("The latest version of {0} may "
                   "be downloaded from".format(_markup(project_name)))
        ret.append(".UR {0}".format(_markup(url)))
        ret.append(".UE")
    return ret


def get_footer(data):
    """
    Return a manual page footer based on the data returned from
    get_manpage_data_from_distribution().  Used only by the old build_manpage
    module.
    """
    return "\n".join(_get_footer_lines(data)) + "\n"


# This is already considered an API, and seems like a valid scenario:
# https://github.com/pypa/pipx/blob/fd6650bcaeca3088/scripts/generate_man.py

class Manpage(object):
    # pylint: disable=too-many-instance-attributes
    def __init__(self, parser, _data=None, format='pretty'):
        """
        Manual page abstraction.  Generates, with the help of formater, a manual
        page by __str__() method.  Please avoid using the private _data
        argument (see https://github.com/praiskup/argparse-manpage/issues/7),
        instead override the `self.<ATTRIBUTE>` when needed.
        """
        self.prog = parser.prog
        self.parser = parser
        self.format = format
        self._data = _data or {}
        self._match_texts = []
        if not getattr(parser, '_manpage', None):
            self.parser._manpage = []

        self.manfile = self._data.get("manfile")
        if self.manfile:
            if len(self._data) > 1:
                raise ValueError("manfile set, so no other key is allowed")
            return

        self.formatter = self.parser._get_formatter()
        self.mf = _ManpageFormatter(self.prog, self.formatter, format=self.format)
        self.synopsis = self.parser.format_usage().split(':')[-1].split()

        self.date = self._data.get("date")
        if not self.date:
            builddate = datetime.datetime.utcfromtimestamp(
                int(os.environ.get('SOURCE_DATE_EPOCH', time.time()))
            )
            self.date = builddate.strftime('%Y-%m-%d')

        self.source = self._data.get("project_name")
        if not self.source:
            self.source = self.prog

        version = self._data.get("version")
        if version:
            self.source += " " + str(version)

        self.manual = self._data.get("manual_title")
        if not self.manual:
            self.manual = "Generated Python Manual"

        self.section = self._data.get("manual_section")
        if not self.section:
            self.section = 1

        self.description = self._data.get("description")

        include = self._data.get("include")
        if include is not None:
            self.parse_include(include)

    def format_text(self, text):
        # Wrap by parser formatter and convert to manpage format
        return self.mf.format_text(self.formatter._format_text(text)).strip('\n')

    def __str__(self):
        if self.manfile:
            with open(self.manfile) as fd:
                return fd.read()

        lines = []

        # Header
        # per man (7) man-pages: .TH title section date source manual
        header = '.TH {title} "{section}" "{date}" "{source}" "{manual}"'
        lines.append(header.format(
            title=_markup(self.prog.upper()),
            section=self.section,
            date=_markup(self.date),
            source=_markup(self.source),
            manual=_markup(self.manual),
        ))

        # Name
        lines.append('.SH NAME')
        line = self.prog

        description = None
        if getattr(self.parser, 'man_short_description', None):
            # Let's keep this undocumented.  There's a way to specify this in
            # setup.cfg: 'description'
            description = self.parser.man_short_description
        if self.description:
            description = self.description
        if description:
            line += " - " + description
        lines.append(_markup(line))

        # Synopsis
        synopsis_section = self.get_extra_section("synopsis")
        if self.synopsis or synopsis_section:
            lines.append('.SH SYNOPSIS')
            if synopsis_section:
                lines.append(synopsis_section["content"])
            else:
                lines.append('.B {}'.format(_markup(self.synopsis[0])))
                lines.append(' '.join(self.synopsis[1:]))

        extra_description = None
        description_section = self.get_extra_section("description")
        if description_section:
            extra_description = description_section["content"]
        lines.extend(self.mf.format_parser(self.parser, extra_description=extra_description))

        comments_section = self.get_extra_section("comments")
        if self.parser.epilog or comments_section:
            lines.append("")
            lines.append('.SH COMMENTS')
            if comments_section:
                lines.append(comments_section["content"])
            else:
                lines.append(self.format_text(self.parser.epilog))

        # Additional sections
        for section in self.parser._manpage: # pylint: disable=protected-access
            if section["heading"] not in SPECIAL_MANPAGE_SECTIONS:
                lines.append('.SH {}'.format(section['heading'].upper()))
                lines.append(section['content'])

        lines.append("")
        lines.extend(self.mf.format_footer(self._data))

        # Finally add --include sections that match text in the page
        final_lines = []
        for line in lines:
            final_lines.append(line)
            for match in self._match_texts:
                if re.search(match['match_text'], line):
                    final_lines.append(match['content'])

        return "\n".join(final_lines).strip("\n") + "\n"

    def get_extra_section(self, heading):
        """
        Return supplementary section for the `Manpage` (created with
        `--include`), or `None`
        """
        for section in self.parser._manpage: # pylint: disable=protected-access
            if section["heading"] == heading:
                return section
        return None

    def add_section(self, heading, position, content):
        """
        Add a supplementary section to a `Manpage`
        """
        # Sections that need special treatment
        heading = heading.lower()
        if heading in ("author", "distribution"):
            if heading == "author":
                self._data['authors'] = [content]
            elif heading == "distribution":
                self._data['url'] = content
        section = self.get_extra_section(heading)
        if section is None:
            section = {"heading": heading, "content": ""}
            self.parser._manpage.append(section) # pylint: disable=protected-access
        if position == '<':
            section["content"] = content + section["content"]
        elif position == '=':
            section["content"] = content
        elif position == '>':
            section["content"] += content
        else:
            raise ValueError("invalid position " + position)

    def parse_include(self, file):
        """
        Parse include file and add its contents to the man page
        """
        def get_section(lines, n):
            for i, line in enumerate(lines[n:]):
                if re.match(r'[\[/]', line):
                    return n + i, lines[n:n + i]
            return len(lines), lines[n:]

        with open(file) as f:
            lines = f.readlines()
            i = 0
            while i < len(lines):
                # Parse a header line
                m = re.match(r"/([^/]+)/$", lines[i])
                if m:
                    match_text = m.group(1)
                    i, section_lines = get_section(lines, i + 1)
                    self._match_texts.append({"match_text": match_text, "content": "".join(section_lines).strip()})
                else:
                    m = re.match(r"\[([<=>])?([^\]]+)\]$", lines[i])
                    if m:
                        position = m.group(1) or '<'
                        heading = m.group(2).upper()
                        if heading == "NAME":
                            raise ValueError("Invalid include section " + heading)
                        i, section_lines = get_section(lines, i + 1)
                        self.add_section(heading, position, "".join(section_lines).strip())
                    else:
                        raise ValueError("Invalid or missing section header in include file %s:\n%s" % (file, lines[i]))


def underline(text):
    """
    Wrap text with \fI for underlined text
    """
    return r'\fI\,{0}\/\fR'.format(_markup(text))


def bold(text):
    """ Wrap text by "bold" groff tags """
    return r"\fB{0}\fR".format(_markup(text))


def quoted(text):
    """ Wrap by single-quotes """
    return "'{0}'".format(text)


class _ManpageFormatter(HelpFormatter):
    def __init__(self, prog, old_formatter, format):
        super(HelpFormatter, self).__init__()
        self._prog = prog
        self.of = old_formatter
        assert format in ("pretty", "single-commands-section")
        self.format = format

    @staticmethod
    def _get_aliases_str(aliases):
        if not aliases:
            return ""
        return " (" + ", ".join(aliases) + ")"

    def _format_action_invocation(self, action):
        if not action.option_strings:
            metavar, = self._metavar_formatter(action, action.dest)(1)
            return bold(metavar)

        parts = []

        # if the Optional doesn't take a value, format is:
        #    -s, --long
        if action.nargs == 0:
            parts.extend(map(bold, action.option_strings))

        # if the Optional takes a value, format is:
        #    -s ARGS, --long ARGS
        else:
            default = action.dest.upper()
            args_string = self._format_args(action, default)
            for option_string in action.option_strings:
                parts.append('{} {}'.format(bold(option_string),
                                            underline(args_string)))
        return ', '.join(parts)

    def _format_parser(self, parser, subcommand=None, aliases=None, help=None, extra_description=None):
        # The parser "tree" looks like
        # ----------------------------
        # Parser -> [ActionGroup, ActionGroup, ..]
        # Group -> [Action, Action, ..]
        # Action -> Option
        # Action -> Subparsers
        # Subparser -> [Parser, Parser, ..] So called "choices".

        lines = []
        if subcommand:
            if self.format == "pretty":
                lines.append("")
                # start a new section for each command
                first_line = ".SH COMMAND"
                first_line += " " + underline(quoted(subcommand))
            elif self.format == "single-commands-section":
                # do not start a new section, start subsection of COMMANDS instead
                first_line = ".SS"
                first_line += " " + bold(subcommand + self._get_aliases_str(aliases))
            lines.append(first_line)

            if help:
                if self.format == "pretty":
                    # help is printed on top in the list of commands already
                    pass
                elif self.format == "single-commands-section":
                    # print help
                    lines.append(help)
                    lines.append("")

            lines.append(self.format_text(parser.format_usage()))

        if parser.description or extra_description:
            if subcommand:
                lines.append("")
            else:
                lines.append(".SH DESCRIPTION")

            if extra_description:
                lines.append(extra_description)
            if parser.description:
                lines.append(self.format_text(parser.description))

        is_subsequent_ag = True
        for group in parser._action_groups:
            ag_lines = self._format_action_group(group, subcommand)
            if not ag_lines:
                continue
            if is_subsequent_ag:
                lines.append("")
            lines.extend(ag_lines)
            is_subsequent_ag = True

        return lines

    def format_parser(self, parser, extra_description=None):
        """
        Return lines Groff formatted text for given parser
        """
        return self._format_parser(parser, extra_description=extra_description)

    def _format_action(self, action):
        parts = []
        parts.append('.TP')

        action_header = self._format_action_invocation(action)
        parts.append(action_header)

        # if there was help for the action, add lines of help text
        if action.help:
            help_text = self.of._format_text(self._expand_help(action)).strip('\n')
            parts.append(self.format_text(help_text))

        return parts

    def _format_ag_subcommands(self, actions, prog):
        lines = []

        for action in actions:
            if getattr(action, 'help', None) == SUPPRESS:
                continue
            lines.append('.TP')
            lines.append(bold(prog) + ' ' + underline(action.dest))
            if hasattr(action, 'help'):
                lines.append(self.format_text(action.help))

        return '\n'.join(lines)

    def _format_subparsers(self, action_group, action, subcommand=None):
        lines = []

        if subcommand:
            if self.format == "pretty":
                # start a new section for each command
                lines.append('.SH')
                title = action_group.title.upper()
                title += " " + underline(quoted(subcommand))
                lines.append(title)
            elif self.format == "single-commands-section":
                # do not start a new section, append subsections to the COMMANDS section
                pass
        else:
            # start a new section on top-level
            lines.append('.SH')
            title = action_group.title.upper()
            lines.append(title)

        if self.format == "pretty":
            # print list of subcommands
            lines.append(self._format_ag_subcommands(action._choices_actions,
                         subcommand or self._prog))
        elif self.format == "single-commands-section":
            # skip printing list of subcommands
            pass

        # gather (sub-)command aliases
        command_aliases = {}
        command_aliases_names = set()
        for name, command in action._name_parser_map.items():
            if command not in command_aliases:
                command_aliases[command] = []
            else:
                command_aliases[command].append(name)
                command_aliases_names.add(name)

        command_help = {}
        for i in action._choices_actions:
            command_help[i.dest] = i.help

        for name, choice in action.choices.items():
            if name in command_aliases_names:
                # don't print aliased commands multiple times
                continue
            new_subcommand = "{} {}".format(subcommand or self._prog, name)
            aliases = command_aliases[choice]
            help = command_help.get(name, None)
            if help == SUPPRESS:
                # don't print hidden commands
                continue
            lines.extend(self._format_parser(choice, new_subcommand, aliases, help))

        return lines

    def _format_action_group(self, action_group, subcommand=None):
        # Parser consists of these action_groups:
        # - positional arguments (no group_actions)
        # - ungrouped options
        # - group 1 options
        # - group 2 options
        # - ...
        # - subparsers

        content = []
        some_action = False
        for action in action_group._group_actions:
            if action.help == SUPPRESS:
                continue

            if isinstance(action, _SubParsersAction):
                return self._format_subparsers(action_group, action,
                                               subcommand)

            if '--help' in action.option_strings:
                # TODO: put out some man page comment ..
                continue

            if some_action:
                # Separate actions
                content.append("")

            some_action = True
            content.extend(self._format_action(action))

        # We don't print empty argument groups.
        if not some_action:
            return []

        title = action_group.title

        group_names = DEFAULT_GROUP_NAMES
        if subcommand:
            if self.format == "pretty":
                pass
            elif self.format == "single-commands-section":
                group_names = DEFAULT_GROUP_NAMES_SUBCOMMANDS

        for replace_with, defaults in group_names.items():
            if title in defaults:
                title = replace_with

        if subcommand:
            if self.format == "pretty":
                title = title.upper() if title else ""
                if title:
                    title += " " + underline(quoted(subcommand))
                title = [] if not title else [".SH " + title]
            elif self.format == "single-commands-section":
                title = [] if not title else [title]
        else:
            title = title.upper() if title else ""
            title = [] if not title else [".SH " + title]

        description = []
        if action_group.description:
            description.append(self.format_text(action_group.description))
            description.append("")

        if subcommand:
            if self.format == "pretty":
                # don't indent the whole content of a subcommand
                pass
            elif self.format == "single-commands-section":
                # indent the whole content of a subcommand
                content = [".RS 7"] + content + [".RE"] + [""]

        return title + description + content

    @staticmethod
    def format_text(text):
        """
        Format a block of text as it was a single line in set of other lines
        (e.g. no trailing newline).
        """
        return _markup(text.strip('\n'))

    @staticmethod
    def format_footer(data):
        """
        Get lines for footer.
        """
        return _get_footer_lines(data)



"""
Compat hacks for argparse-manpage's build_manpages module
"""

# pylint: disable=unused-import,deprecated-module,raise-missing-from

try:
    from setuptools import Command
    from setuptools.command.build_py import build_py
    from setuptools.command.install import install
except ImportError:
    try:
        from distutils.core import Command
        from distutils.command.build_py import build_py
        from distutils.command.install import install
    except ImportError as orig_err:
        raise ImportError(
            "To use the 'build_manpages' tool on Python 3.12+, "
            "you need to install 'setuptools'."
        )

# A separate try-except block from the one above.  This is more likely to fail
# than the above setuptools part and we would start using distutils just
# beacause of the exception.
try:
    from setuptools.errors import OptionError as DistutilsOptionError
except ImportError:
    from distutils.errors import DistutilsOptionError



# -*- coding: utf-8 -*-

"""build_manpage command -- Generate man page from setup()"""

import datetime
import optparse
import argparse
import os
import time
import warnings

from distutils.core import Command
from distutils.errors import DistutilsOptionError

from argparse_manpage.manpage import (
    get_manpage_data_from_distribution,
    get_footer,
)

from argparse_manpage.tooling import (
    get_parser_from_file,
    get_parser_from_module,
    write_to_filename,
)

warnings.warn(
    "The 'build_manpage' module will be removed in the next argparse-manpage "
    "version v5.  Please migrate to 'build_manpages'.", DeprecationWarning,
)

class ManPageWriter(object):
    _parser = None
    _command = None
    _type = None

    def __init__(self, parser, values):
        self._parser = parser
        self.values = values
        self._today = datetime.datetime.utcfromtimestamp(
            int(os.environ.get('SOURCE_DATE_EPOCH', time.time()))
        )

        if isinstance(parser, argparse.ArgumentParser):
            self._type = 'argparse'
            if parser.formatter_class == argparse.HelpFormatter:
                # Hack for issue #36, to have reproducible manual page content
                # regardless the terminal window size.  Long term we should avoid
                # using the built-in usage formatter, and generate our own.
                parser.formatter_class = \
                        lambda prog: argparse.HelpFormatter(prog, width=78)

        else:
            self._parser.formatter = ManPageFormatter()
            self._parser.formatter.set_parser(self._parser)

    def _markup(self, txt):
        return txt.replace('-', '\\-')

    def _write_header(self):
        version = self.values["version"]
        prog = self.values["prog"]
        ret = []

        ret.append('.TH %s 1 %s "%s v.%s"\n' % (self._markup(prog),
                                      self._today.strftime('%Y\\-%m\\-%d'), prog, version))

        description = self.values.get("description")
        if description:
            name = self._markup('%s - %s' % (prog, description.splitlines()[0]))
        else:
            name = self._markup(prog)

        ret.append('.SH NAME\n%s\n' % name)
        if getattr(self._parser, 'format_usage', None):
            synopsis = self._parser.format_usage()
        else:
            synopsis = self._parser.get_usage()

        if synopsis:
            synopsis = synopsis.replace('%s ' % prog, '')
            ret.append('.SH SYNOPSIS\n.B %s\n%s\n' % (self._markup(prog),
                                                      synopsis))
        long_desc = self.values.get("long_description", "")
        if long_desc:
            ret.append('.SH DESCRIPTION\n%s\n' % self._markup(long_desc))
        return ''.join(ret)

    def _write_options(self, action_name=None, parser=None):
        if not parser:
            parser = self._parser

        if not action_name:
            ret = ['.SH OPTIONS\n']
        else:
            ret = ['.SH OPTIONS ' + action_name.upper() + '\n']

        ret.append(parser.format_option_help())
        if self._type != 'argparse':
            return ''.join(ret)

        subparsers_actions = [
            action for action in parser._actions
                if isinstance(action, argparse._SubParsersAction)]

        for subparser_action in subparsers_actions:
            for name, obj in subparser_action.choices.items():
                if action_name:
                    an = action_name + " " + name
                else:
                    an = name
                ret.append(self._write_options(an, obj))

        return ''.join(ret)

    def _write_seealso(self, text):
        ret = []
        ret.append('.SH "SEE ALSO"\n')

        for i in text:
            name, sect = i.split(":")

            if len(ret) > 1:
                ret.append(',\n')

            ret.append('.BR %s (%s)' % (name, sect))

        return ''.join(ret)

    def write(self, filename, seealso=None):
        manpage = []
        manpage.append(self._write_header())
        manpage.append(self._write_options())
        manpage.append(get_footer(self.values))
        if seealso:
            manpage.append(self._write_seealso(seealso))
        write_to_filename(''.join(manpage), filename)


class build_manpage(Command):

    description = 'Generate man page from setup().'

    user_options = [
        ('output=', 'O', 'output file'),
        ('parser=', None, 'module path to optparser (e.g. mymod:func'),
        ('parser-file=', None, 'file to the parser module'),
        ('file-and-object=', None, 'import parser object from file, e.g. "bin/blah.py:fooparser"'),
        ('seealso=', None, 'list of manpages to put into the SEE ALSO section (e.g. bash:1)')
        ]

    def initialize_options(self):
        self.output = None
        self.parser = None
        self.seealso = None
        self.parser_file = None
        self.file_and_object = None

    def _get_parser_from_module(self):
        mod_name, func_name = self.parser.split(':')
        if self.parser_file:
            # The 'modname' is entirely ignored in this case.  This is a design
            # issue from 516ca12512979ab8e1a45f24e502a9cd1331f284.  But we keep
            # the 'build_manpage' for compatibility reasons.
            return get_parser_from_file(self.parser_file, func_name,
                                            'function')
        return get_parser_from_module(mod_name, func_name, 'function')

    def finalize_options(self):
        if self.output is None:
            raise DistutilsOptionError('\'output\' option is required')
        if self.parser is None and self.file_and_object is None:
            raise DistutilsOptionError('\'parser\' or \'file-and-object\' option is required')

        self.ensure_string_list('seealso')

        if self.file_and_object:
            filename, objname = self.file_and_object.split(':')
            self._parser = get_parser_from_file(filename, objname)
        else:
            self._parser = self._get_parser_from_module()


    def run(self):
        self.announce('Writing man page %s' % self.output)

        data = {}
        get_manpage_data_from_distribution(self.distribution, data)

        # the format is actually unused
        mpw = ManPageWriter(self._parser, data)
        mpw.write(self.output, seealso=self.seealso)


class ManPageFormatter(optparse.HelpFormatter):

    def __init__(self,
                 indent_increment=2,
                 max_help_position=24,
                 width=None,
                 short_first=1):
        optparse.HelpFormatter.__init__(self, indent_increment,
                                        max_help_position, width, short_first)

    def _markup(self, txt):
        return txt.replace('-', '\\-')

    def format_usage(self, usage):
        return self._markup(usage)

    def format_heading(self, heading):
        if self.level == 0:
            return ''
        return '.TP\n%s\n' % self._markup(heading.upper())

    def format_option(self, option):
        result = []
        opts = self.option_strings[option]
        result.append('.TP\n.B %s\n' % self._markup(opts))
        if option.help:
            help_text = '%s\n' % self._markup(self.expand_default(option))
            result.append(help_text)
        return ''.join(result)



"""
build_manpages command -- generate set of manual pages by the setup()
command.
"""

import os
import shutil

try:
    import tomllib
    from tomllib import TOMLDecodeError
except ImportError:
    try:
        import tomli as tomllib
        from tomli import TOMLDecodeError
    except ImportError:
        import toml as tomllib
        from toml import TomlDecodeError as TOMLDecodeError

from argparse_manpage.compat import ConfigParser, NoSectionError
from argparse_manpage.tooling import get_parser, write_to_filename
from argparse_manpage.manpage import (
    Manpage,
    MANPAGE_DATA_ATTRS,
    get_manpage_data_from_distribution,
)

from .compat import (
    build_py,
    Command,
    DistutilsOptionError,
    install,
)

DEFAULT_CMD_NAME = 'build_manpages'

def parse_manpages_spec(string):
    manpages_data = {}
    for spec in string.strip().split('\n'):
        manpagedata = {}
        output = True
        for option in spec.split(':'):
            if output:
                outputfile = option
                output = False
                continue

            oname, ovalue = option.split('=')

            if oname == 'function' or oname == 'object':
                assert(not 'objtype' in manpagedata)
                manpagedata['objtype'] = oname
                manpagedata['objname'] = ovalue

            elif oname == 'pyfile' or oname == 'module':
                assert(not 'import_type' in manpagedata)
                manpagedata['import_type'] = oname
                manpagedata['import_from'] = ovalue
                if oname == 'pyfile':
                    manpagedata['prog'] = os.path.basename(ovalue)

            elif oname == 'format':
                assert(not 'format' in manpagedata)
                manpagedata[oname] = ovalue

            elif oname == 'author':
                manpagedata.setdefault("authors", []).append(ovalue)

            elif oname in MANPAGE_DATA_ATTRS and oname != "authors":
                assert(not oname in manpagedata)
                manpagedata[oname] = ovalue

            else:
                raise ValueError("Unknown manpage configuration option: {}".format(oname))

        manpages_data[outputfile] = manpagedata

    return manpages_data

def get_pyproject_settings():
    """Parse and handle errors of a toml configuration file."""
    try:
        with open("pyproject.toml", mode="r") as fp:
            content = tomllib.loads(fp.read())
    except TOMLDecodeError:
        return None

    try:
        value = content["tool"][DEFAULT_CMD_NAME]["manpages"]
        if isinstance(value, list):
            value = "\n".join(value)
        return str(value)
    except KeyError:
        return None

class build_manpages(Command):
    description = 'Generate set of man pages from setup().'
    user_options = [
        ('manpages=', 'O', 'list man pages specifications'),
    ]

    def initialize_options(self):
        self.manpages = None


    def finalize_options(self):
        manpages = self.manpages or get_pyproject_settings()
        if not manpages:
            raise DistutilsOptionError('\'manpages\' option is required')
        self.manpages_data = parse_manpages_spec(manpages)

        # if a value wasn't set in setup.cfg, use the value from setup.py
        for page, data in self.manpages_data.items():
            get_manpage_data_from_distribution(self.distribution, data)

    def run(self):
        for page, data in self.manpages_data.items():
            if data.get('manfile'):
                print ("using pre-written " + page)
                return
            print ("generating " + page)
            parser = get_parser(data['import_type'], data['import_from'], data['objname'], data['objtype'], data.get('prog', None))
            format = data.get('format', 'pretty')
            if format in ('pretty', 'single-commands-section'):
                manpage = Manpage(parser, format=format, _data=data)
                write_to_filename(str(manpage), page)
            elif format == 'old':
                # TODO: drop the "old" format support, and stop depending on ManPageWriter
                # pylint: disable=import-outside-toplevel
                from .build_manpage import ManPageWriter
                mw = ManPageWriter(parser, data)
                mw.write(page)
            else:
                raise ValueError("Unknown format: {}".format(format))


def get_build_py_cmd(command=build_py):
    """
    Override the default 'setup.py build_py' command with one that automatically
    generates manual pages.  By default we use an overridden
    'setuptools.command.build_py' (class).  If your project already uses an
    overridden class, specify the optional 'command=YourCommandClass`.
    """
    class _build_manpages_build_py(command):
        def run(self):
            self.run_command(DEFAULT_CMD_NAME)
            command.run(self)

    return _build_manpages_build_py


def get_install_cmd(command=install):
    """
    Override the default 'setup.py install' command with one that automatically
    installs the manual pages generated by `build_manpages`, see
    'build_manpages.get_build_py_cmd()'.  By default we use the
    'setuptools.command.install' class as the base.  If you already use an
    such an overridden class, set the optional 'command=YourCommandClass'.
    """
    class _build_manpages_install(command):
        def install_manual_pages(self):
            """
            Additional logic for installing the generated manual pages
            """
            config = ConfigParser()
            config.read('setup.cfg')
            try:
                spec = config.get(DEFAULT_CMD_NAME, 'manpages')
            except NoSectionError:
                spec = get_pyproject_settings()
            if spec is None:
                raise ValueError("'manpage' configuration not found in setup.cfg or pyproject.toml")

            data = parse_manpages_spec(spec)

            mandir = os.path.join(self.install_data, 'man/man1')
            if not os.path.exists(mandir):
                os.makedirs(mandir)
            for key, _ in data.items():
                print ('installing {0}'.format(key))
                shutil.copy(key, mandir)

        def run(self):
            command.run(self)
            self.install_manual_pages()

    return _build_manpages_install



# Copyright (C) 2017 Red Hat, Inc.

# Python 2.7 hack.  Without this, 'from build_manpages.build_manpage' attempts
# to import 'build_manpages.build_manpages.build_manpage' because of our
# unfortunate file naming.
from __future__ import absolute_import

import argparse

from argparse_manpage.tooling import get_parser, write_to_filename
from argparse_manpage.manpage import MANPAGE_DATA_ATTRS, Manpage


description = """
Build manual page from Python's argparse.ArgumentParser object.
""".strip()

ap = argparse.ArgumentParser(
    prog='argparse-manpage',
    description=description,
)

src_group = ap.add_mutually_exclusive_group(required=True)
src_group.add_argument(
    "--module",
    help="search the OBJECT/FUNCTION in MODULE"
)

src_group.add_argument(
    "--pyfile",
    help="search the OBJECT/FUNCTION in FILE"
)

obj_group = ap.add_mutually_exclusive_group(required=True)
obj_group.add_argument(
    "--function",
    help="call FUNCTION from MODULE/FILE to obtain ArgumentParser object",
)
obj_group.add_argument(
    "--object",
    help="obtain ArgumentParser OBJECT from FUNCTION (to get argparse object) from MODULE or FILE",
)


ap.add_argument("--project-name", help="Name of the project the documented program is part of.")
ap.add_argument("--prog", help="Substitutes %%prog in ArgumentParser's usage.")
ap.add_argument("--version", help=(
    "Version of the program, will be visible in the "
    "manual page footer."))
ap.add_argument("--description", metavar="TEXT", help=(
    "description of the program, used in the NAME section, after the "
    "leading 'name - ' part, see man (7) man-pages for more info"))
ap.add_argument("--long-description", metavar="TEXT", help=argparse.SUPPRESS)
ap.add_argument("--author", action="append", dest="authors", metavar="[AUTHOR]",
                help="Author of the program. Can be specified multiple times.")
ap.add_argument("--author-email", action="append", dest="authors",
                help=argparse.SUPPRESS)
ap.add_argument("--url", help="Link to project's homepage")
ap.add_argument("--format", default="pretty", choices=("pretty", "single-commands-section"),
                help="Format of the generated man page. Defaults to 'pretty'.")
ap.add_argument("--output", dest='outfile', default='-',
                help="Output file. Defaults to stdout.")
ap.add_argument("--manual-section", help=(
    "Section of the manual, by default 1.  See man (7) man-pages for more "
    "info about existing sections."))
ap.add_argument("--manual-title", help=(
    "The title of the manual, by default \"Generated Python Manual\". "
    "See man (7) man-pages for more instructions."))
ap.add_argument("--include", metavar="FILE", help=(
    "File that contains extra material for the man page."))
ap.add_argument("--manfile", metavar="FILE", help=(
    "File containing a complete man page."))


def args_to_manpage_data(args):
    data = {}
    for attr in MANPAGE_DATA_ATTRS:
        value = getattr(args, attr)
        data[attr] = value
    return data


def main():
    args = ap.parse_args()

    import_type = 'pyfile'
    import_from = args.pyfile
    if args.module:
        import_type = 'module'
        import_from = args.module

    obj_type = 'object'
    obj_name = args.object
    if args.function:
        obj_type = 'function'
        obj_name = args.function

    parser = get_parser(import_type, import_from, obj_name, obj_type, prog=args.prog)
    data = args_to_manpage_data(args)
    manpage = Manpage(parser, format=args.format, _data=data)
    write_to_filename(str(manpage), args.outfile)



"""
Compatibility hacks for the argparse-manpage project.
"""

import sys

# Drop once Python 2.7 is dropped
# pylint: disable=unused-import
try:
    from configparser import ConfigParser, NoSectionError
except ImportError:
    from ConfigParser import SafeConfigParser as ConfigParser, NoSectionError  # type: ignore

if sys.version_info < (3, 0):
    import imp  # pylint: disable=deprecated-module
    def load_py_file(filename):
        """ Small wrapper having the same call arg list as runpy.run_path() """
        return imp.load_source("argparse_manpage_loaded_file", filename)
else:
    from runpy import run_path as load_py_file

def get_module_object(module_or_dict, objname, objtype):
    """
    Get OBJNAME from a given MODULE (or dict, if loaded using runpy.run_path(),
    but call the object first if OBJTYPE is 'function'.
    """
    obj = None
    if isinstance(module_or_dict, dict):
        obj = module_or_dict[objname]
    else:
        obj = getattr(module_or_dict, objname)
    if objtype != 'object':
        obj = obj()
    return obj


def load_file_as_module(filename):
    """
    Load a given python filename as a dict (runpy on Python 3) or as a module
    (imp module, Python 2).  Note that 'runpy.run_path()' doesn't work correctly
    with Python 2.7 (the imported object doesn't see it's own globals/imported
    modules), and 'imp' module is deprecated for modern Python 3.
    """

    # We used to call 'runpy.run_path()' here, but that did not work correctly
    # with Python 2.7 where the imported object did not see it's own
    # globals/imported modules (including the 'argparse' module).
    return load_py_file(filename)



"""
A tooling helpers for the argparse-manpage project.
"""

import importlib
import os
import sys

from .compat import load_file_as_module, get_module_object


def _environ_hack():
    os.environ['BUILD_MANPAGES_RUNNING'] = 'TRUE'


def get_parser_from_module(module, objname, objtype='object', prog=None):
    """
    Read the given module and return the requested object from there.
    """
    _environ_hack()
    # We need to set argv[0] to properly so argparse returns appropriate "usage"
    # strings.  Like "usage: argparse-manpage [-h] ...", instead of
    # "usage: setup.py ...".
    backup_argv = sys.argv
    if prog:
        sys.argv = [prog]

    mod = importlib.import_module(module)
    obj = get_module_object(mod, objname, objtype)

    # Restore callee's argv
    sys.argv = backup_argv
    return obj


def get_parser_from_file(filename, objname, objtype='object', prog=None):
    """
    Load the given filename as a module and return the requested object from
    there.
    """
    _environ_hack()
    # We need to set argv[0] to properly so argparse returns appropriate "usage"
    # strings.  Like "usage: argparse-manpage [-h] ...", instead of
    # "usage: setup.py ...".
    backup_argv = sys.argv
    if prog:
        sys.argv = [prog]
    else:
        sys.argv = [os.path.basename(filename)]

    # Get the ArgumentParser object
    module_loaded = load_file_as_module(filename)
    obj = get_module_object(module_loaded, objname, objtype)

    # Restore callee's argv
    sys.argv = backup_argv
    return obj


def get_parser(import_type, import_from, objname, objtype, prog=None):
    """
    Load a function or object from a given file or module.
    """
    if import_type == 'pyfile':
        return get_parser_from_file(import_from, objname, objtype, prog=prog)
    return get_parser_from_module(import_from, objname, objtype, prog=prog)


def write_to_filename(text, filename):
    """
    Write given text into a filename at once.  Pre-create the parent directory
    if it doesn't exist yet.  Print to stdout if filename == '-'.
    """
    filename = filename if filename != '-' else '/dev/stdout'
    dirname = os.path.dirname(filename)
    if dirname and not os.path.exists(dirname):
        os.makedirs(dirname)
    with open(filename, 'w') as stream:
        stream.write(text)
